"""Dump strongly-typed containers into loosely-typed objects."""

from dataclasses import asdict, fields, is_dataclass
from enum import Enum
from typing import Any, Mapping, Sequence

from .errors import SerializeError
from .overrides import get_override
from .typedefs import MISSING, NamedTupleType


def _filter_keep(value: Any, keep_always: bool) -> bool:
    """Determine whether we should keep a value in serialized output."""
    if keep_always:
        return True
    return value is not MISSING


def _identity(value: Any) -> Any:
    """Simple identity function, returning original value."""
    return value


def dump(obj: Any, convert_missing_to_none: bool = False) -> Any:
    """Serialize an object into a lesser-typed form.

    Parameters:
        obj: the object that you would like to serialize.
        convert_missing_to_none: flag indicating whether or not we should
            retain a missing property's key and convert its value to `None`.
            This will keep all keys when serializing. Useful if you want to
            keep all column names and assign value of `None` to missing
            columns.

    Returns:
        A serialized form of `obj`.

    Raises:
        sedelicacy.SerializeError: raised for any unhandled error

    Notes:
        If serializing a dataclass and `transform_dump` metadata exists in a
        `dataclass`'s '`dataclasses.field`, its value is assumed to be function
        whose result is serialized before being passed recursively down the
        chain.

        Serialize from an instance of `a` -> an instance of `b`:
            | `dataclass` -> `Dict`
            | `NamedTuple` -> `Dict`
            | `Enum` -> enum value
            | `str` -> `str`
            | `Sequence` -> `List`
            | `Mapping` -> `Dict`
            | `MISSING` -> `None` (if `convert_missing_to_none` is `True`)
            | `MISSING` filtered out (if `convert_missing_to_none` is `False`)
            | `Anything else` -> `itself`
    """
    # pylint: disable=too-many-return-statements
    try:
        if is_dataclass(obj):
            custom_dump = {
                f.name: get_override(f.metadata.get("serdelicacy"))
                for f in fields(obj)
            }
            return {
                key: dump(__value_converted, convert_missing_to_none)
                for key, value in asdict(obj).items()
                if _filter_keep(
                    (
                        __value_converted := custom_dump[key].transform_dump(
                            value
                        )
                    ),
                    convert_missing_to_none,
                )
            }
        if isinstance(obj, NamedTupleType):
            return {
                key: dump(value, convert_missing_to_none)
                for key, value in obj._asdict().items()
                if _filter_keep(value, convert_missing_to_none)
            }
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, str):
            return obj
        if isinstance(obj, Sequence):
            return [
                dump(value, convert_missing_to_none)
                for value in obj
                if _filter_keep(value, convert_missing_to_none)
            ]
        if isinstance(obj, Mapping):
            return {
                dump(key, convert_missing_to_none): dump(
                    value, convert_missing_to_none
                )
                for key, value in obj.items()
                if _filter_keep(value, convert_missing_to_none)
            }
        if convert_missing_to_none and (obj is MISSING):
            return None
        return obj
    except Exception as error:
        raise SerializeError(f"Error serializing {repr(object)}") from error



"""Serdelicacy overrides for dataclasses."""

from dataclasses import dataclass, field
from typing import Any, Callable, NoReturn, TypeVar, Union

T = TypeVar("T")  # pylint: disable=invalid-name


def _noreturn(value: Any) -> NoReturn:  # pylint: disable=unused-argument
    """Simple 1 argument function that returns nothing."""


def _id(value: T) -> T:
    """Identity function."""
    return value


@dataclass(frozen=True)
class Override:
    """User-defined overrides for serde with `datacalsses.dataclass`

    Should be passed into the `metadata` argument to `dataclasses.field` via
    the "serdelicacy" key.

    Parameters:
        validate: a function that either returns `True` on positive validation
            / `False` on non-validation, or returns nothing at all and instead
            relies on the raising of exceptions to indicate whether validation
            passed for failed.
        transform_load: a function that, when deserializing, is evaluated on an
            object before the object is recursively examined.
        transform_postload: a function that, when deserializing, is evaluated
            on an object after the object has been recursively examined. When
            possible, the `transform_load` should be preferred over
            `transform_postload`, but there are situations where
            `transform_postload` is useful.
        transform_dump: a function that, when serializing a dataclass, is
            called on its value before it is recursively serialized.

    Example:

        The following example shows how a function, returning `True`/`False`,
        is passed to `serdelicacy.load` through the `metadata` parameter to
        `dataclasses.field`.

        .. code-block:: python

            from dataclasses import dataclass, field
            import serdelicacy
            from serdelicacy import Override

            BOOK_RAW = {"author": "sam"}

            @dataclass
            class Book:
                author: str = field(
                    metadata={"serdelicacy": Override(validate=str.istitle)}
                )

            BOOK = serdelicacy.load(BOOK_RAW, Book)

        This example should raise a `serdelicacy.DeserializeError`
    """

    validate: Union[
        Callable[[Any], NoReturn],
        Callable[[Any], bool],
        Callable[[Any], None],
    ] = field(default=_noreturn)
    transform_load: Callable[[Any], Any] = field(default=_id)
    transform_postload: Callable[[Any], Any] = field(default=_id)
    transform_dump: Callable[[Any], Any] = field(default=_id)


DEFAULT_OVERRIDE = Override()


def get_override(value: Any) -> Override:
    """Perform validation for override value."""
    if value is None:
        return DEFAULT_OVERRIDE
    if isinstance(value, Override):
        return value
    raise TypeError(
        "dataclasses field error for metadata key 'serdelicacy': value "
        f"{repr(value)} is not the correct type. Expected {repr(Override)}"
    )



"""Goal: load and dump data into strongly-typed data structures.

1. "Deserialize" unstructured Python types into structured, type-hinted Python
   types (dataclasses.dataclass, typing.NamedTuples).
2. "Serialize" structured, type-hinted Python objects into unstructured Python
   types (eg, the reverse).

The top-level module contains `serdelicacy`'s public API and is directly
importable from `serdelicacy`. If you choose to import names from anything
outside of this module, do so at your own risk. Code organization in the
package submodules is subject to change at any time is not reflected in this
library's semantic versioning strategy.
"""

from .deserialize import load
from .errors import DeserializeError, SerdeError, SerializeError
from .overrides import Override
from .serialize import dump
from .typedefs import OptionalProperty, get, is_missing
